"""
API for TensorFlow usage.
"""
import itertools
from typing import Any, Callable, List, Optional, Sequence, Tuple

import numpy as np
import tensorflow as tf  # type: ignore

from targetran._check import (
    _check_shear_input,
    _check_translate_input,
    _check_crop_input,
    _check_input_range,
)
from targetran._tf_functional import (
    _tf_convert,
    _tf_cast_to_int,
    _tf_round_to_int,
    _tf_resize_image,
    _tf_pad_image,
    _tf_gather_image,
)
from targetran._transform import (
    _AffineDependency,
    _affine_transform,
    _flip_left_right,
    _flip_up_down,
    _rotate,
    _shear,
    _translate,
    _get_crop_inputs,
    _get_random_size_fractions,
    _crop,
    _resize,
    _get_flip_left_right_mats,
    _get_flip_up_down_mats,
    _get_rotate_mats,
    _get_shear_mats,
    _get_translate_mats,
)
from targetran._typing import T
from targetran.utils import Interpolation


def load_tf_image(image_path: str) -> tf.Tensor:
    return tf.image.decode_image(
        tf.io.read_file(image_path), channels=3, expand_animations=False
    )


def to_tf(
        image_seq: Sequence[T],
        bboxes_seq: Sequence[T],
        labels_seq: Sequence[T],
        image_seq_is_paths: bool = False,
) -> Tuple[Sequence[tf.Tensor], Sequence[tf.Tensor], Sequence[tf.Tensor]]:
    """
    Convert seqs to TF (eager) tensor seqs.
    """
    tuples: List[Tuple[tf.Tensor, tf.Tensor, tf.Tensor]] = []
    for image, bboxes, labels in itertools.zip_longest(  # type: ignore
            image_seq, bboxes_seq, labels_seq, fillvalue=[]
    ):
        tuples.append(
            (image if image_seq_is_paths else _tf_convert(image),
             tf.reshape(_tf_convert(bboxes), (-1, 4)),
             _tf_convert(labels))
        )
    tf_image_seq, tf_bboxes_seq, tf_labels_seq = tuple(zip(*tuples))

    if image_seq_is_paths:
        tf_image_seq = tf.convert_to_tensor(tf_image_seq, tf.string)

    return tf_image_seq, tf_bboxes_seq, tf_labels_seq


def to_tf_dataset(
        image_seq: Sequence[T],
        bboxes_seq: Sequence[T],
        labels_seq: Sequence[T],
        image_seq_is_paths: bool = False,
) -> tf.data.Dataset:
    tf_image_seq, tf_bboxes_seq, tf_labels_seq = to_tf(
        image_seq, bboxes_seq, labels_seq, image_seq_is_paths
    )

    if image_seq_is_paths:
        ds_image = tf.data.Dataset.from_tensor_slices(tf_image_seq)
        ds_image = ds_image.map(load_tf_image)
    else:
        ds_image = tf.data.Dataset.from_tensor_slices(tf.ragged.stack(tf_image_seq))

    # Tensors of different shapes can be included in a TF Dataset
    # as ragged-tensors.
    ds = tf.data.Dataset.zip((
        ds_image,
        tf.data.Dataset.from_tensor_slices(tf.ragged.stack(tf_bboxes_seq)),
        tf.data.Dataset.from_tensor_slices(tf.ragged.stack(tf_labels_seq))
    ))
    # However, our transformations expect normal tensors, so the ragged-tensors
    # have to be first converted back to tensors during mapping. Therefore,
    # the whole point of using ragged-tensors is ONLY for building a Dataset...
    # Note that the label ragged-tensors are of rank-0, so they are implicitly
    # converted to tensors during mapping. Strange TF Dataset behaviour...
    ds = ds.map(
        lambda i, b, l: (
            i if isinstance(i, tf.Tensor) else i.to_tensor(),
            b if isinstance(b, tf.Tensor) else b.to_tensor(),
            l
        )
    )
    return ds


def to_keras_cv_dict(
        ds: tf.data.Dataset,
        batch_size: Optional[int] = None,
        drop_remainder: bool = True,
) -> tf.data.Dataset:
    import keras_cv  # type: ignore

    ds = ds.map(lambda i, b, l: (i, {"boxes": b, "classes": l}))
    if batch_size:
        ds = ds.ragged_batch(batch_size=batch_size, drop_remainder=drop_remainder)

    ds = ds.map(lambda i, d: {"images": i, "bounding_boxes": d})
    return ds


def to_keras_cv_model_input(
        ds: tf.data.Dataset,
        max_num_bboxes: Optional[int] = None,
        fill_value: int = -1,
) -> tf.data.Dataset:
    import keras_cv

    return ds.map(
        lambda d: (
            d["images"],
            keras_cv.bounding_box.to_dense(
                d["bounding_boxes"],
                max_boxes=max_num_bboxes,
                default_value=fill_value
            )
        )
    )


def _tf_get_affine_dependency() -> _AffineDependency:
    return _AffineDependency(
        _tf_convert, tf.shape, tf.reshape, tf.expand_dims, tf.squeeze,
        _tf_pad_image, tf.range, _tf_cast_to_int, _tf_round_to_int,
        tf.repeat, tf.tile, tf.ones_like, tf.stack, tf.concat, tf.matmul,
        tf.clip_by_value, tf.math.floor, tf.math.ceil, tf.transpose,
        _tf_gather_image, tf.identity, tf.reduce_max, tf.reduce_min,
        tf.logical_and, tf.boolean_mask
    )


def _tf_affine_transform(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        image_dest_tran_mat: tf.Tensor,
        bboxes_tran_mat: tf.Tensor,
        interpolation: Interpolation,
        fill_value: float
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _affine_transform(
        image, bboxes, labels, image_dest_tran_mat, bboxes_tran_mat,
        interpolation, fill_value, _tf_get_affine_dependency()
    )


def tf_flip_left_right(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _flip_left_right(
        image, bboxes, labels, _tf_get_affine_dependency()
    )


def tf_flip_up_down(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _flip_up_down(
        image, bboxes, labels, _tf_get_affine_dependency()
    )


def tf_rotate(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        angle_deg: float,
        interpolation: Interpolation = Interpolation.BILINEAR,
        fill_value: float = 0.0
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _rotate(
        image, bboxes, labels, _tf_convert(angle_deg), tf.cos, tf.sin,
        interpolation, fill_value, _tf_get_affine_dependency()
    )


def tf_shear(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        angle_deg: float,
        interpolation: Interpolation = Interpolation.BILINEAR,
        fill_value: float = 0.0,
        _check_input: bool = True
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    if _check_input:
        _check_shear_input(angle_deg)
    return _shear(
        image, bboxes, labels, _tf_convert(angle_deg), tf.tan,
        interpolation, fill_value, _tf_get_affine_dependency()
    )


def tf_translate(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        translate_height: int,
        translate_width: int,
        interpolation: Interpolation = Interpolation.BILINEAR,
        fill_value: float = 0.0,
        _check_input: bool = True
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    if _check_input:
        _check_translate_input(
            image.get_shape(), translate_height, translate_width
        )
    return _translate(
        image, bboxes, labels,
        _tf_convert(translate_height), _tf_convert(translate_width),
        interpolation, fill_value, _tf_get_affine_dependency()
    )


def _tf_get_crop_inputs(
        image_height: int,
        image_width: int,
        height_fraction_range: Tuple[float, float],
        width_fraction_range: Tuple[float, float],
        rand_fn: Callable[..., tf.Tensor]
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    return _get_crop_inputs(
        image_height, image_width, height_fraction_range, width_fraction_range,
        rand_fn, _tf_convert, _tf_round_to_int
    )


def tf_crop(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        offset_height: int,
        offset_width: int,
        crop_height: int,
        crop_width: int,
        _check_input: bool = True
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    if _check_input:
        _check_crop_input(image.get_shape(), offset_height, offset_width)
    return _crop(
        image, bboxes, labels,
        _tf_convert(offset_height), _tf_convert(offset_width),
        _tf_convert(crop_height), _tf_convert(crop_width),
        _tf_convert, tf.shape, tf.reshape, tf.concat,
        tf.logical_and, tf.squeeze, tf.clip_by_value, tf.boolean_mask
    )


def tf_resize(
        image: tf.Tensor,
        bboxes: tf.Tensor,
        labels: tf.Tensor,
        dest_size: Tuple[int, int],
        interpolation: Interpolation = Interpolation.BILINEAR
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    return _resize(
        image, bboxes, labels, dest_size, interpolation,
        _tf_convert, tf.shape, tf.reshape, _tf_resize_image, tf.concat
    )


class TFRandomTransform:

    def __init__(
            self,
            tf_fn: Callable[..., Tuple[tf.Tensor, tf.Tensor, tf.Tensor]],
            probability: float,
            seed: Optional[int],
            name: str,
            is_affine: bool
    ) -> None:
        self._tf_fn = tf_fn
        self.probability = probability
        self.seed = seed if seed is not None else np.random.randint(1e6)  # type: ignore
        self._rng = tf.random.Generator.from_seed(self.seed)
        self.name = name
        self.is_affine = is_affine

    def _rand_fn(self, shape: Sequence[int] = ()) -> tf.Tensor:
        return self._rng.uniform(shape=shape)

    def _get_mats(  # type: ignore[empty-body]
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        pass

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        # Make sure inputs are in the needed format.
        image = _tf_convert(image)
        bboxes = _tf_convert(bboxes)
        labels = _tf_convert(labels)

        if self._rand_fn() < _tf_convert(self.probability):
            return self._tf_fn(image, bboxes, labels, *args, **kwargs)
        return image, bboxes, labels


def _get_random_indices(
        rng: tf.random.Generator,
        num_indices: int,
        num_selected_indices: int,
        selected_probabilities: Optional[List[float]] = None,
) -> tf.Tensor:
    """
    Roughly mimicking Numpy choice for getting indices, without replacement.
    The indices always start from 0.

    References:
    https://github.com/tensorflow/tensorflow/issues/9260#issuecomment-437875125
    """
    probs = selected_probabilities if selected_probabilities \
        else tf.ones(num_indices)
    logits = tf.math.log(probs)
    z = -tf.math.log(-tf.math.log(  # pylint: disable=invalid-unary-operand-type
        rng.uniform(tf.shape(logits), 0, 1)
    ))
    _, indices = tf.math.top_k(logits + z, num_selected_indices, sorted=False)
    return indices


class TFCombineAffine(TFRandomTransform):

    def __init__(
            self,
            transforms: Sequence[TFRandomTransform],
            num_selected_transforms: Optional[int] = None,
            selected_probabilities: Optional[List[float]] = None,
            keep_order: bool = False,
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 1.0,
            seed: Optional[int] = None
    ) -> None:
        not_affine_trans = list(filter(lambda t: not t.is_affine, transforms))
        if not_affine_trans:
            raise AssertionError(
                f"Non-affine transforms cannot be included in TFCombineAffine: "
                f"{[t.name for t in not_affine_trans]}"
            )
        if num_selected_transforms and selected_probabilities:
            if len(selected_probabilities) != len(transforms):
                raise ValueError(
                    "Number of items in selected_probabilities should be "
                    "the same as the number of items in transforms."
                )
        super().__init__(
            _tf_affine_transform, probability, seed, "TFCombineAffine", True
        )
        self._transforms = transforms
        self._num_selected_transforms = num_selected_transforms
        self._selected_probabilities = selected_probabilities
        self._keep_order = keep_order
        self._interpolation = interpolation
        self._fill_value = fill_value

    def _get_mats(
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        image_dest_tran_mats, bboxes_tran_mats, probs = tuple(zip(
            *[(*t._get_mats(image, rand_fn), t.probability)
              for i, t in enumerate(self._transforms)]
        ))

        if self._num_selected_transforms is not None:
            if self._num_selected_transforms == 0:
                return tf.eye(3), tf.eye(3)

            indices = _get_random_indices(
                self._rng,
                len(self._transforms),
                self._num_selected_transforms,
                self._selected_probabilities
            )
        else:
            # It looks like in graph mode we cannot check if a tensor is empty
            # and then do short circuit. Therefore, to prevent empty indices
            # which tf.gather could not handle, an identity matrix is appended.
            image_dest_tran_mats += (tf.eye(3),)
            bboxes_tran_mats += (tf.eye(3),)
            probs += (1.0,)

            conditions = rand_fn() < probs
            indices = tf.boolean_mask(
                tf.range(len(probs), dtype=tf.int32), conditions
            )

        if self._keep_order:
            indices = tf.sort(indices)
        else:
            indices = tf.random.shuffle(indices, seed=self.seed)

        image_dest_tran_mats = tf.gather(  # pylint: disable=no-value-for-parameter
            image_dest_tran_mats, indices
        )
        bboxes_tran_mats = tf.gather(  # pylint: disable=no-value-for-parameter
            bboxes_tran_mats, indices
        )

        def matmul(a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
            return tf.matmul(a, b)

        image_dest_tran_mat = tf.scan(matmul, image_dest_tran_mats)[-1]
        # Note the reversed order for the bboxes tran matrices.
        bboxes_tran_mat = tf.scan(matmul, bboxes_tran_mats[::-1])[-1]
        return image_dest_tran_mat, bboxes_tran_mat

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        image_dest_tran_mat, bboxes_tran_mat = self._get_mats(
            image, self._rand_fn
        )
        return super().__call__(
            image, bboxes, labels, image_dest_tran_mat, bboxes_tran_mat,
            self._interpolation, self._fill_value
        )


class TFRandomFlipLeftRight(TFRandomTransform):

    def __init__(
            self,
            probability: float = 0.5,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(
            tf_flip_left_right, probability, seed, "TFRandomFlipLeftRight", True
        )

    def _get_mats(
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        return _get_flip_left_right_mats(_tf_convert)

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(image, bboxes, labels)


class TFRandomFlipUpDown(TFRandomTransform):

    def __init__(
            self,
            probability: float = 0.5,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(
            tf_flip_up_down, probability, seed, "TFRandomFlipUpDown", True
        )

    def _get_mats(
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        return _get_flip_up_down_mats(_tf_convert)

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(image, bboxes, labels)


class TFRandomRotate(TFRandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-15.0, 15.0),
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(angle_deg_range, None, "angle_deg_range")
        super().__init__(tf_rotate, probability, seed, "TFRandomRotate", True)
        self.angle_deg_range = angle_deg_range
        self.interpolation = interpolation
        self.fill_value = fill_value

    def _get_angle_deg(self, rand_fn: Callable[..., tf.Tensor]) -> tf.Tensor:
        return _tf_convert(self.angle_deg_range[1] - self.angle_deg_range[0]) \
               * rand_fn() + _tf_convert(self.angle_deg_range[0])

    def _get_mats(
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        return _get_rotate_mats(
            self._get_angle_deg(rand_fn), _tf_convert, tf.cos, tf.sin
        )

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(
            image, bboxes, labels, self._get_angle_deg(self._rand_fn),
            self.interpolation, self.fill_value
        )


class TFRandomShear(TFRandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-10.0, 10.0),
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(angle_deg_range, (-90.0, 90.0), "angle_deg_range")
        super().__init__(tf_shear, probability, seed, "TFRandomShear", True)
        self.angle_deg_range = angle_deg_range
        self.interpolation = interpolation
        self.fill_value = fill_value

    def _get_angle_deg(self, rand_fn: Callable[..., tf.Tensor]) -> tf.Tensor:
        return _tf_convert(self.angle_deg_range[1] - self.angle_deg_range[0]) \
               * rand_fn() + _tf_convert(self.angle_deg_range[0])

    def _get_mats(
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        return _get_shear_mats(
            self._get_angle_deg(rand_fn), _tf_convert, tf.tan
        )

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return super().__call__(
            image, bboxes, labels, self._get_angle_deg(self._rand_fn),
            self.interpolation, self.fill_value, False
        )


class TFRandomTranslate(TFRandomTransform):

    def __init__(
            self,
            translate_height_fraction_range: Tuple[float, float] = (-0.1, 0.1),
            translate_width_fraction_range: Tuple[float, float] = (-0.1, 0.1),
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(
            translate_height_fraction_range, (-1.0, 1.0),
            "translate_height_fraction_range"
        )
        _check_input_range(
            translate_width_fraction_range, (-1.0, 1.0),
            "translate_width_fraction_range"
        )
        super().__init__(
            tf_translate, probability, seed, "TFRandomTranslate", True
        )
        self.translate_height_fraction_range = translate_height_fraction_range
        self.translate_width_fraction_range = translate_width_fraction_range
        self.interpolation = interpolation
        self.fill_value = fill_value

    def _get_translate_height_and_width(
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        height_fraction, width_fraction = _get_random_size_fractions(
            self.translate_height_fraction_range,
            self.translate_width_fraction_range,
            rand_fn, _tf_convert
        )
        translate_height = _tf_convert(tf.shape(image)[0]) * height_fraction
        translate_width = _tf_convert(tf.shape(image)[1]) * width_fraction
        return translate_height, translate_width

    def _get_mats(
            self,
            image: tf.Tensor,
            rand_fn: Callable[..., tf.Tensor]
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        translate_height, translate_width = \
            self._get_translate_height_and_width(image, rand_fn)
        return _get_translate_mats(
            translate_height, translate_width, _tf_convert
        )

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        translate_height, translate_width = \
            self._get_translate_height_and_width(image, self._rand_fn)
        return super().__call__(
            image, bboxes, labels, translate_height, translate_width,
            self.interpolation, self.fill_value, False
        )


class TFRandomCrop(TFRandomTransform):

    def __init__(
            self,
            crop_height_fraction_range: Tuple[float, float] = (0.8, 0.9),
            crop_width_fraction_range: Tuple[float, float] = (0.8, 0.9),
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(
            crop_height_fraction_range, (0.0, 1.0), "crop_height_fraction_range"
        )
        _check_input_range(
            crop_width_fraction_range, (0.0, 1.0), "crop_width_fraction_range"
        )
        super().__init__(tf_crop, probability, seed, "TFRandomCrop", False)
        self.crop_height_fraction_range = crop_height_fraction_range
        self.crop_width_fraction_range = crop_width_fraction_range

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        offset_height, offset_width, crop_height, crop_width = \
            _tf_get_crop_inputs(
                tf.shape(image)[0], tf.shape(image)[1],
                self.crop_height_fraction_range,
                self.crop_width_fraction_range,
                self._rand_fn
            )

        return super().__call__(
            image, bboxes, labels,
            offset_height, offset_width, crop_height, crop_width, False
        )


class TFResize:

    def __init__(
            self,
            dest_size: Tuple[int, int],
            interpolation: Interpolation = Interpolation.BILINEAR
    ) -> None:
        self.dest_size = dest_size
        self.interpolation = interpolation
        self.name = "TFResize"
        self.is_affine = False

    def __call__(
            self,
            image: tf.Tensor,
            bboxes: tf.Tensor,
            labels: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        return tf_resize(image, bboxes, labels, self.dest_size, self.interpolation)



#!/usr/bin/env python3
"""
PyTorch Dataset test.
"""
from typing import Optional, Sequence, Tuple

import numpy as np
import numpy.typing
from torch.utils.data import Dataset, DataLoader

import targetran.np
from targetran.utils import Compose, collate_fn

NDAnyArray = np.typing.NDArray[np.float_]


def make_np_data() -> Tuple[Sequence[NDAnyArray],
                            Sequence[NDAnyArray],
                            Sequence[NDAnyArray]]:
    image_seq = [np.random.rand(480, 512, 3) for _ in range(3)]

    bboxes_seq = [
        np.array([
            [214, 223, 10, 11],
            [345, 230, 21, 9],
        ]),
        np.array([]),
        np.array([
            [104, 151, 22, 10],
            [99, 132, 20, 15],
            [340, 220, 31, 12],
        ]),
    ]

    labels_seq = [
        np.array([0, 1]),
        np.array([]),
        np.array([2, 3, 0]),
    ]

    return image_seq, bboxes_seq, labels_seq


class PTDataset(Dataset):

    def __init__(
            self,
            image_seq: Sequence[NDAnyArray],
            bboxes_seq: Sequence[NDAnyArray],
            labels_seq: Sequence[NDAnyArray],
            transforms: Optional[Compose]
    ) -> None:
        self.image_seq = image_seq
        self.bboxes_seq = bboxes_seq
        self.labels_seq = labels_seq
        self.transforms = transforms

    def __len__(self) -> int:
        return len(self.image_seq)

    def __getitem__(
            self,
            idx: int
    ) -> Tuple[NDAnyArray, NDAnyArray, NDAnyArray]:
        if self.transforms:
            return self.transforms(
                self.image_seq[idx],
                self.bboxes_seq[idx],
                self.labels_seq[idx]
            )
        return (
            self.image_seq[idx],
            self.bboxes_seq[idx],
            self.labels_seq[idx]
        )


def main() -> None:
    image_seq, bboxes_seq, labels_seq = make_np_data()

    transforms = Compose([
        targetran.np.RandomRotate(probability=1.0),
        targetran.np.RandomShear(probability=1.0),
        targetran.np.RandomTranslate(probability=1.0),
        targetran.np.RandomFlipUpDown(probability=1.0),
        targetran.np.RandomFlipLeftRight(probability=1.0),
        targetran.np.RandomCrop(probability=1.0),
        targetran.np.Resize((256, 256)),
    ])

    print("-------- Random transform --------")

    ds = PTDataset(image_seq, bboxes_seq, labels_seq, transforms)

    for example in ds:
        image, bboxes, labels = example
        print(f"transformed image shape: {image.shape}")
        print(f"transformed bboxes shape: {bboxes.shape}")
        print(f"transformed bboxes: {bboxes.tolist()}")
        print(f"transformed labels shape: {labels.shape}")
        print(f"transformed labels: {labels.tolist()}")
        print("=========")

    print("-------- Batching --------")

    data_loader = DataLoader(ds, batch_size=2, collate_fn=collate_fn)

    for batch in data_loader:
        image_tuple, bboxes_tuple, labels_tuple = batch
        print(f"transformed image batch size: {len(image_tuple)}")
        print(f"transformed bboxes batch size: {len(bboxes_tuple)}")
        print(f"transformed labels batch size: {len(labels_tuple)}")
        print(f"image shapes: {[i.shape for i in image_tuple]}")
        print(f"bboxes shapes: {[b.shape for b in bboxes_tuple]}")
        print(f"labels shapes: {[l.shape for l in labels_tuple]}")
        print("=========")


if __name__ == "__main__":
    main()



"""
Typing utilities.
"""
from typing import Any

import numpy as np
import numpy.typing

ArrayLike = np.typing.ArrayLike
NDAnyArray = np.typing.NDArray[Any]
NDBoolArray = np.typing.NDArray[np.bool_]
NDFloatArray = np.typing.NDArray[np.float_]
NDIntArray = np.typing.NDArray[np.int_]

# T is treated semantically as "NDAnyArray or tf.Tensor" in this library.
# However, there is currently no way to express this accurately in the
# typing system. Hence, it is assumed here that tf.Tensor is compatible
# with NDAnyArray, which is kind of true. Hope there will be a better way.
T = NDAnyArray



from ._tf import (
    to_tf as to_tf,
    to_tf_dataset as to_tf_dataset,
    to_keras_cv_dict as to_keras_cv_dict,
    to_keras_cv_model_input as to_keras_cv_model_input,
    tf_flip_left_right as tf_flip_left_right,
    tf_flip_up_down as tf_flip_up_down,
    tf_rotate as tf_rotate,
    tf_shear as tf_shear,
    tf_translate as tf_translate,
    tf_crop as tf_crop,
    tf_resize as tf_resize,
    TFCombineAffine as TFCombineAffine,
    TFRandomTransform as TFRandomTransform,
    TFRandomFlipLeftRight as TFRandomFlipLeftRight,
    TFRandomFlipUpDown as TFRandomFlipUpDown,
    TFRandomRotate as TFRandomRotate,
    TFRandomShear as TFRandomShear,
    TFRandomTranslate as TFRandomTranslate,
    TFRandomCrop as TFRandomCrop,
    TFResize as TFResize,
)



"""
API for NumPy usage.
"""
from typing import Any, Callable, List, Optional, Sequence, Tuple

import numpy as np

from targetran._check import (
    _check_shear_input,
    _check_translate_input,
    _check_crop_input,
    _check_input_range,
)
from targetran._np_functional import (
    _np_convert,
    _np_range,
    _np_cast_to_int,
    _np_round_to_int,
    _np_resize_image,
    _np_boolean_mask,
    _np_logical_and,
    _np_pad_image,
    _np_gather_image,
)
from targetran._transform import (
    _AffineDependency,
    _affine_transform,
    _flip_left_right,
    _flip_up_down,
    _rotate,
    _shear,
    _translate,
    _get_crop_inputs,
    _get_random_size_fractions,
    _crop,
    _resize,
    _get_flip_left_right_mats,
    _get_flip_up_down_mats,
    _get_rotate_mats,
    _get_shear_mats,
    _get_translate_mats,
)
from targetran._typing import NDFloatArray, NDIntArray
from targetran.utils import Interpolation


def _np_get_affine_dependency() -> _AffineDependency:
    return _AffineDependency(
        _np_convert, np.shape, np.reshape, np.expand_dims, np.squeeze,
        _np_pad_image, _np_range, _np_cast_to_int, _np_round_to_int, np.repeat,
        np.tile,  # type: ignore
        np.ones_like, np.stack, np.concatenate, np.matmul,
        np.clip, np.floor, np.ceil, np.transpose, _np_gather_image,
        np.copy, np.max, np.min,
        _np_logical_and, _np_boolean_mask
    )


def _np_affine_transform(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray,
        image_dest_tran_mat: NDFloatArray,
        bboxes_tran_mat: NDFloatArray,
        interpolation: Interpolation,
        fill_value: float
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    return _affine_transform(
        image, bboxes, labels, image_dest_tran_mat, bboxes_tran_mat,
        interpolation, fill_value, _np_get_affine_dependency()
    )


def flip_left_right(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    return _flip_left_right(
        image, bboxes, labels, _np_get_affine_dependency()
    )


def flip_up_down(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    return _flip_up_down(
        image, bboxes, labels, _np_get_affine_dependency()
    )


def rotate(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray,
        angle_deg: float,
        interpolation: Interpolation = Interpolation.BILINEAR,
        fill_value: float = 0.0
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    return _rotate(
        image, bboxes, labels, _np_convert(angle_deg), np.cos, np.sin,
        interpolation, fill_value, _np_get_affine_dependency()
    )


def shear(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray,
        angle_deg: float,
        interpolation: Interpolation = Interpolation.BILINEAR,
        fill_value: float = 0.0,
        _check_input: bool = True
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    if _check_input:
        _check_shear_input(angle_deg)
    return _shear(
        image, bboxes, labels, _np_convert(angle_deg), np.tan,
        interpolation, fill_value, _np_get_affine_dependency()
    )


def translate(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray,
        translate_height: int,
        translate_width: int,
        interpolation: Interpolation = Interpolation.BILINEAR,
        fill_value: float = 0.0,
        _check_input: bool = True
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    if _check_input:
        _check_translate_input(image.shape, translate_height, translate_width)
    return _translate(
        image, bboxes, labels,
        _np_convert(translate_height), _np_convert(translate_width),
        interpolation, fill_value, _np_get_affine_dependency()
    )


def _np_get_crop_inputs(
        image_height: int,
        image_width: int,
        height_fraction_range: Tuple[float, float],
        width_fraction_range: Tuple[float, float],
        rand_fn: Callable[..., NDFloatArray]
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray, NDFloatArray]:
    return _get_crop_inputs(
        image_height, image_width, height_fraction_range, width_fraction_range,
        rand_fn, _np_convert, _np_round_to_int
    )


def crop(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray,
        offset_height: int,
        offset_width: int,
        crop_height: int,
        crop_width: int,
        _check_input: bool = True
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    if _check_input:
        _check_crop_input(image.shape, offset_height, offset_width)
    return _crop(
        image, bboxes, labels,
        _np_convert(offset_height), _np_convert(offset_width),
        _np_convert(crop_height), _np_convert(crop_width),
        _np_convert, np.shape, np.reshape, np.concatenate,
        _np_logical_and, np.squeeze, np.clip, _np_boolean_mask
    )


def resize(
        image: NDFloatArray,
        bboxes: NDFloatArray,
        labels: NDFloatArray,
        dest_size: Tuple[int, int],
        interpolation: Interpolation = Interpolation.BILINEAR
) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
    return _resize(
        image, bboxes, labels, dest_size, interpolation,
        _np_convert, np.shape, np.reshape, _np_resize_image, np.concatenate
    )


class RandomTransform:

    def __init__(
            self,
            np_fn: Callable[..., Tuple[NDFloatArray, NDFloatArray, NDFloatArray]],
            probability: float,
            seed: Optional[int],
            name: str,
            is_affine: bool
    ) -> None:
        self._np_fn = np_fn
        self.probability = probability
        self.seed = seed
        self._rng = np.random.default_rng(seed=self.seed)
        self.name = name
        self.is_affine = is_affine

    def _rand_fn(self, shape: Sequence[int] = ()) -> NDFloatArray:
        return self._rng.random(shape)

    def _get_mats(  # type: ignore[empty-body]
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDFloatArray, NDFloatArray]:
        pass

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        # Make sure inputs are in the needed format.
        image = _np_convert(image)
        bboxes = _np_convert(bboxes)
        labels = _np_convert(labels)

        if self._rand_fn() < self.probability:
            return self._np_fn(image, bboxes, labels, *args, **kwargs)
        return image, bboxes, labels


class CombineAffine(RandomTransform):

    def __init__(
            self,
            transforms: Sequence[RandomTransform],
            num_selected_transforms: Optional[int] = None,
            selected_probabilities: Optional[List[float]] = None,
            keep_order: bool = False,
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 1.0,
            seed: Optional[int] = None
    ) -> None:
        not_affine_trans = list(filter(lambda t: not t.is_affine, transforms))
        if not_affine_trans:
            raise AssertionError(
                f"Non-affine transforms cannot be included in CombineAffine: "
                f"{[t.name for t in not_affine_trans]}"
            )
        if num_selected_transforms and selected_probabilities:
            if len(selected_probabilities) != len(transforms):
                raise ValueError(
                    "Number of items in selected_probabilities should be "
                    "the same as the number of items in transforms."
                )
        super().__init__(
            _np_affine_transform, probability, seed, "CombineAffine", True
        )
        self._transforms = transforms
        self._num_selected_transforms = num_selected_transforms
        self._selected_probabilities = selected_probabilities
        self._keep_order = keep_order
        self._interpolation = interpolation
        self._fill_value = fill_value

    def _get_mats(
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDFloatArray, NDFloatArray]:
        image_dest_tran_mats, bboxes_tran_mats, probs = tuple(zip(
            *[(*t._get_mats(image, rand_fn), t.probability)
              for i, t in enumerate(self._transforms)]
        ))

        if self._num_selected_transforms is not None:
            indices = self._rng.choice(
                len(self._transforms),
                self._num_selected_transforms,
                replace=False, p=self._selected_probabilities
            )
        else:
            conditions = rand_fn() < probs
            indices = np.arange(len(probs), dtype=np.int32)[conditions]  # type: ignore

        if len(indices) > 1:
            if self._keep_order:
                indices.sort()
            else:
                self._rng.shuffle(indices)

            indices = indices.tolist()
            image_dest_tran_mats = np.take(image_dest_tran_mats, indices, 0)
            bboxes_tran_mats = np.take(bboxes_tran_mats, indices, 0)

            image_dest_tran_mat = np.linalg.multi_dot(image_dest_tran_mats)
            # Note the reversed order for the bboxes tran matrices.
            bboxes_tran_mat = np.linalg.multi_dot(bboxes_tran_mats[::-1])

        elif len(indices) == 1:
            image_dest_tran_mat = image_dest_tran_mats[0]
            bboxes_tran_mat = bboxes_tran_mats[0]
        else:
            # This happens when conditions are all False, or
            # num_selected_transforms is 0.
            assert len(indices) == 0
            image_dest_tran_mat = np.identity(3)
            bboxes_tran_mat = np.identity(3)

        return image_dest_tran_mat, bboxes_tran_mat

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        image_dest_tran_mat, bboxes_tran_mat = self._get_mats(
            image, self._rand_fn
        )
        return super().__call__(
            image, bboxes, labels, image_dest_tran_mat, bboxes_tran_mat,
            self._interpolation, self._fill_value
        )


class RandomFlipLeftRight(RandomTransform):

    def __init__(
            self,
            probability: float = 0.5,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(
            flip_left_right, probability, seed, "RandomFlipLeftRight", True
        )

    def _get_mats(
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDFloatArray, NDFloatArray]:
        return _get_flip_left_right_mats(_np_convert)

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        return super().__call__(image, bboxes, labels)


class RandomFlipUpDown(RandomTransform):

    def __init__(
            self,
            probability: float = 0.5,
            seed: Optional[int] = None
    ) -> None:
        super().__init__(
            flip_up_down, probability, seed, "RandomFlipUpDown", True
        )

    def _get_mats(
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDFloatArray, NDFloatArray]:
        return _get_flip_up_down_mats(_np_convert)

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        return super().__call__(image, bboxes, labels)


class RandomRotate(RandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-15.0, 15.0),
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(angle_deg_range, None, "angle_deg_range")
        super().__init__(rotate, probability, seed, "RandomRotate", True)
        self.angle_deg_range: NDFloatArray = np.array(angle_deg_range)
        self.interpolation = interpolation
        self.fill_value = fill_value

    def _get_angle_deg(
            self,
            rand_fn: Callable[..., NDFloatArray]
    ) -> NDFloatArray:
        return (  # type: ignore
            self.angle_deg_range[1] - self.angle_deg_range[0]
            * rand_fn() + self.angle_deg_range[0]
        )

    def _get_mats(
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDFloatArray, NDFloatArray]:
        return _get_rotate_mats(
            self._get_angle_deg(rand_fn), _np_convert, np.cos, np.sin
        )

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        return super().__call__(
            image, bboxes, labels, self._get_angle_deg(self._rand_fn),
            self.interpolation, self.fill_value
        )


class RandomShear(RandomTransform):

    def __init__(
            self,
            angle_deg_range: Tuple[float, float] = (-10.0, 10.0),
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(angle_deg_range, (-90.0, 90.0), "angle_deg_range")
        super().__init__(shear, probability, seed, "RandomShear", True)
        self.angle_deg_range: NDFloatArray = np.array(angle_deg_range)
        self.interpolation = interpolation
        self.fill_value = fill_value

    def _get_angle_deg(
            self,
            rand_fn: Callable[..., NDFloatArray]
    ) -> NDFloatArray:
        return (  # type: ignore
            self.angle_deg_range[1] - self.angle_deg_range[0]
            * rand_fn() + self.angle_deg_range[0]
        )

    def _get_mats(
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDFloatArray, NDFloatArray]:
        return _get_shear_mats(self._get_angle_deg(rand_fn), _np_convert, np.tan)

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        return super().__call__(
            image, bboxes, labels, self._get_angle_deg(self._rand_fn),
            self.interpolation, self.fill_value, False
        )


class RandomTranslate(RandomTransform):

    def __init__(
            self,
            translate_height_fraction_range: Tuple[float, float] = (-0.1, 0.1),
            translate_width_fraction_range: Tuple[float, float] = (-0.1, 0.1),
            interpolation: Interpolation = Interpolation.BILINEAR,
            fill_value: float = 0.0,
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(
            translate_height_fraction_range, (-1.0, 1.0),
            "translate_height_fraction_range"
        )
        _check_input_range(
            translate_width_fraction_range, (-1.0, 1.0),
            "translate_width_fraction_range"
        )
        super().__init__(translate, probability, seed, "RandomTranslate", True)
        self.translate_height_fraction_range = translate_height_fraction_range
        self.translate_width_fraction_range = translate_width_fraction_range
        self.interpolation = interpolation
        self.fill_value = fill_value

    def _get_translate_height_and_width(
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDIntArray, NDIntArray]:
        height_fraction, width_fraction = _get_random_size_fractions(
            self.translate_height_fraction_range,
            self.translate_width_fraction_range,
            rand_fn, _np_convert
        )
        translate_height = _np_round_to_int(
            np.shape(image)[0] * height_fraction
        )
        translate_width = _np_round_to_int(
            np.shape(image)[1] * width_fraction
        )
        return translate_height, translate_width

    def _get_mats(
            self,
            image: NDFloatArray,
            rand_fn: Callable[..., NDFloatArray]
    ) -> Tuple[NDFloatArray, NDFloatArray]:
        translate_height, translate_width = \
            self._get_translate_height_and_width(image, rand_fn)
        return _get_translate_mats(
            translate_height, translate_width, _np_convert
        )

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        translate_height, translate_width = \
            self._get_translate_height_and_width(image, self._rand_fn)
        return super().__call__(
            image, bboxes, labels, translate_height, translate_width,
            self.interpolation, self.fill_value, False
        )


class RandomCrop(RandomTransform):

    def __init__(
            self,
            crop_height_fraction_range: Tuple[float, float] = (0.8, 0.9),
            crop_width_fraction_range: Tuple[float, float] = (0.8, 0.9),
            probability: float = 0.9,
            seed: Optional[int] = None
    ) -> None:
        _check_input_range(
            crop_height_fraction_range, (0.0, 1.0), "crop_height_fraction_range"
        )
        _check_input_range(
            crop_width_fraction_range, (0.0, 1.0), "crop_width_fraction_range"
        )
        super().__init__(crop, probability, seed, "RandomCrop", False)
        self.crop_height_fraction_range = crop_height_fraction_range
        self.crop_width_fraction_range = crop_width_fraction_range

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray,
            *args: Any,
            **kwargs: Any
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        offset_height, offset_width, crop_height, crop_width = \
            _np_get_crop_inputs(
                np.shape(image)[0], np.shape(image)[1],
                self.crop_height_fraction_range,
                self.crop_width_fraction_range,
                self._rand_fn
            )

        return super().__call__(
            image, bboxes, labels,
            offset_height, offset_width, crop_height, crop_width, False
        )


class Resize:

    def __init__(
            self,
            dest_size: Tuple[int, int],
            interpolation: Interpolation = Interpolation.BILINEAR
    ) -> None:
        self.dest_size = dest_size
        self.interpolation = interpolation
        self.name = "Resize"
        self.is_affine = False

    def __call__(
            self,
            image: NDFloatArray,
            bboxes: NDFloatArray,
            labels: NDFloatArray
    ) -> Tuple[NDFloatArray, NDFloatArray, NDFloatArray]:
        return resize(image, bboxes, labels, self.dest_size, self.interpolation)



"""
Image and target transform utilities.
"""
from dataclasses import dataclass
from typing import Callable, List, Sequence, Tuple

import numpy as np

from targetran._typing import T
from targetran.utils import Interpolation


def _sanitise(
        image: T,
        bboxes: T,
        labels: T,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Sequence[int]],
        reshape_fn: Callable[[T, Sequence[int]], T],
) -> Tuple[T, T, T]:
    """
    Try to convert input to the expected format.
    """
    if len(shape_fn(image)) != 3:
        raise ValueError(
            "Input image must have 3 dimensions, i.e., "
            "(height, width, num_channels)."
        )
    image = convert_fn(image)
    bboxes = reshape_fn(convert_fn(bboxes), (-1, 4))
    labels = convert_fn(labels)
    return image, bboxes, labels


@dataclass
class _AffineDependency:
    convert_fn: Callable[..., T]
    shape_fn: Callable[[T], Sequence[int]]
    reshape_fn: Callable[[T, Sequence[int]], T]
    expand_dim_fn: Callable[[T, int], T]
    squeeze_fn: Callable[[T, int], T]
    pad_image_fn: Callable[[T, T], T]
    range_fn: Callable[[int, int, int], T]
    cast_to_int_fn: Callable[[T], T]
    round_to_int_fn: Callable[[T], T]
    repeat_fn: Callable[[T, T], T]
    tile_fn: Callable[[T, T], T]
    ones_like_fn: Callable[[T], T]
    stack_fn: Callable[[List[T], int], T]
    concat_fn: Callable[[List[T], int], T]
    matmul_fn: Callable[[T, T], T]
    clip_fn: Callable[[T, T, T], T]
    floor_fn: Callable[[T], T]
    ceil_fn: Callable[[T], T]
    transpose_fn: Callable[[T], T]
    gather_image_fn: Callable[[T, T], T]
    copy_fn: Callable[[T], T]
    max_fn: Callable[[T, int], T]
    min_fn: Callable[[T, int], T]
    logical_and_fn: Callable[[T, T], T]
    boolean_mask_fn: Callable[[T, T], T]


def _affine_transform(
        image: T,
        bboxes: T,
        labels: T,
        image_dest_tran_mat: T,
        bboxes_tran_mat: T,
        interpolation: Interpolation,
        fill_value: float,
        d: _AffineDependency
) -> Tuple[T, T, T]:
    """
    Original transformation is done w.r.t. the origin (top-left pixel).
    Hence, additional translation is also performed to make the transformation
    w.r.t. the image centre.

    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, d.convert_fn, d.shape_fn, d.reshape_fn
    )

    image_shape = d.shape_fn(image)
    height, width = int(image_shape[0]), int(image_shape[1])
    num_channels = int(image_shape[2])

    # Pad image to provide a fill-value pixel frame for clipping use below.
    fill_image = d.ones_like_fn(image) * fill_value

    pad_offsets = d.convert_fn([1, 1, 1, 1])
    image = d.pad_image_fn(image, pad_offsets)
    fill_image = d.pad_image_fn(fill_image, pad_offsets)

    fill_image = (fill_image - fill_value) * -1.0
    image = image + fill_image

    # References:
    # https://www.kaggle.com/cdeotte/rotation-augmentation-gpu-tpu-0-96

    # Destination indices.
    row_idxes = d.repeat_fn(  # Along y-axis, from top to bottom.
        d.range_fn(0, height, 1),
        d.round_to_int_fn(d.convert_fn([width]))
    )
    col_idxes = d.tile_fn(  # Along x-axis, from left to right.
        d.range_fn(0, width, 1),
        d.round_to_int_fn(d.convert_fn([height]))
    )
    # Note the (col, row) -> (x, y) swapping. Last axis is dummy.
    image_dest_idxes = d.stack_fn(
        [col_idxes, row_idxes, d.ones_like_fn(col_idxes)], 0
    )

    # Transform destination indices, align centre of the idx matrix to
    # that of the image, and clipping. Note that these are floats.
    new_image_dest_idxes = d.matmul_fn(
        image_dest_tran_mat, d.convert_fn(image_dest_idxes)
    )
    image_cen = d.convert_fn([
        [d.convert_fn(width - 1) / 2.0],
        [d.convert_fn(height - 1) / 2.0],
        [0.0]
    ])
    new_image_dest_idxes -= \
        d.matmul_fn(image_dest_tran_mat, image_cen) - image_cen

    clipped_new_image_dest_idxes = d.clip_fn(
        new_image_dest_idxes[:2],
        # Note the extra idx for the padded frame.
        d.convert_fn([[-1], [-1]]),
        d.convert_fn([[width], [height]])
    )

    # Assigning original pixel values to new positions.
    # Note the 1-pixel shift for the padded frame.
    image_orig_idxes = d.concat_fn([
        # Rows.
        clipped_new_image_dest_idxes[1:] + d.convert_fn(1),
        # Columns.
        clipped_new_image_dest_idxes[:1] + d.convert_fn(1)
    ], 0)

    if interpolation == Interpolation.NEAREST:
        image_rounded_orig_idxes = d.round_to_int_fn(image_orig_idxes)
        values = d.gather_image_fn(image, image_rounded_orig_idxes)

    elif interpolation == Interpolation.BILINEAR:
        floor_floor_idxes = d.cast_to_int_fn(d.floor_fn(image_orig_idxes))
        ceil_ceil_idxes = d.cast_to_int_fn(d.ceil_fn(image_orig_idxes))

        floor_ceil_idxes = d.concat_fn(
            [floor_floor_idxes[:1, :], ceil_ceil_idxes[1:, :]], 0
        )
        ceil_floor_idxes = d.concat_fn(
            [ceil_ceil_idxes[:1, :], floor_floor_idxes[1:, :]], 0
        )

        dists: T = image_orig_idxes - d.convert_fn(floor_floor_idxes)
        # Transpose to shape (-1, 2), needed for broadcasting in the later
        # gather-values step. See the return shape of gather_image_fn.
        floor_weights = d.transpose_fn(1.0 - dists)
        ceil_weights: T = 1.0 - floor_weights
        floor_floor_weights = floor_weights[:, :1] * floor_weights[:, 1:]
        floor_ceil_weights = floor_weights[:, :1] * ceil_weights[:, 1:]
        ceil_floor_weights = ceil_weights[:, :1] * floor_weights[:, 1:]
        ceil_ceil_weights = ceil_weights[:, :1] * ceil_weights[:, 1:]

        values = \
            d.gather_image_fn(image, floor_floor_idxes) * floor_floor_weights + \
            d.gather_image_fn(image, floor_ceil_idxes) * floor_ceil_weights + \
            d.gather_image_fn(image, ceil_floor_idxes) * ceil_floor_weights + \
            d.gather_image_fn(image, ceil_ceil_idxes) * ceil_ceil_weights

    else:
        raise ValueError("Unsupported interpolation option.")

    new_image = d.reshape_fn(values, (height, width, num_channels))

    # Transform bboxes.
    top_left_xs = bboxes[:, :1]
    top_left_ys = bboxes[:, 1:2]
    top_right_xs = bboxes[:, :1] + bboxes[:, 2:3] - 1
    top_right_ys = bboxes[:, 1:2]
    bottom_left_xs = d.copy_fn(top_left_xs)
    bottom_left_ys = d.copy_fn(top_left_ys + bboxes[:, 3:] - 1)
    bottom_right_xs = d.copy_fn(top_right_xs)
    bottom_right_ys = d.copy_fn(top_right_ys + bboxes[:, 3:] - 1)

    xs = d.concat_fn(
        [top_left_xs, top_right_xs, bottom_left_xs, bottom_right_xs], 1
    )
    ys = d.concat_fn(
        [top_left_ys, top_right_ys, bottom_left_ys, bottom_right_ys], 1
    )
    bboxes_idxes = d.stack_fn(  # Shape: [num_bboxes, 3, 4].
        [xs, ys, d.ones_like_fn(xs)], 1
    )

    tran_bboxes_idxes = d.matmul_fn(bboxes_tran_mat, bboxes_idxes)
    tran_bboxes_idxes -= d.matmul_fn(bboxes_tran_mat, image_cen) - image_cen

    # New bboxes, defined as the rectangle enclosing the transformed bboxes.
    tran_xs = tran_bboxes_idxes[:, 0, :]  # Shape: [num_bboxes, 4].
    tran_ys = tran_bboxes_idxes[:, 1, :]
    max_xs = d.max_fn(tran_xs, -1)  # Shape: [num_bboxes].
    max_ys = d.max_fn(tran_ys, -1)
    min_xs = d.min_fn(tran_xs, -1)
    min_ys = d.min_fn(tran_ys, -1)

    one = d.cast_to_int_fn(d.convert_fn(1))

    tran_top_left_xs = d.round_to_int_fn(d.expand_dim_fn(min_xs, -1))
    tran_top_left_ys = d.round_to_int_fn(d.expand_dim_fn(min_ys, -1))
    tran_bottom_right_xs = d.round_to_int_fn(d.expand_dim_fn(max_xs, -1))
    tran_bottom_right_ys = d.round_to_int_fn(d.expand_dim_fn(max_ys, -1))
    new_widths: T = tran_bottom_right_xs - tran_top_left_xs + one
    new_heights: T = tran_bottom_right_ys - tran_top_left_ys + one
    tran_bboxes = d.concat_fn([  # Shape: [num_bboxes, 4].
        tran_top_left_xs, tran_top_left_ys, new_widths, new_heights
    ], -1)

    new_xs = tran_bboxes[:, :1]
    new_ys = tran_bboxes[:, 1:2]

    # Filter new bboxes values.
    xcens = new_xs + new_widths // 2
    ycens = new_ys + new_heights // 2
    included = d.squeeze_fn(d.logical_and_fn(
        d.logical_and_fn(xcens >= 0, xcens <= width),
        d.logical_and_fn(ycens >= 0, ycens <= height)
    ), -1)

    # Clip new bboxes values.
    xmaxs = d.clip_fn(
        d.convert_fn(new_xs + new_widths), d.convert_fn(0), d.convert_fn(width)
    )
    ymaxs = d.clip_fn(
        d.convert_fn(new_ys + new_heights), d.convert_fn(0), d.convert_fn(height)
    )
    new_xs = d.clip_fn(
        d.convert_fn(new_xs), d.convert_fn(0), d.convert_fn(width)
    )
    new_ys = d.clip_fn(
        d.convert_fn(new_ys), d.convert_fn(0), d.convert_fn(height)
    )
    new_widths = xmaxs - new_xs
    new_heights = ymaxs - new_ys

    new_bboxes = d.concat_fn([new_xs, new_ys, new_widths, new_heights], 1)
    new_bboxes = d.convert_fn(d.boolean_mask_fn(new_bboxes, included))

    # Filter labels.
    new_labels = d.boolean_mask_fn(labels, included)

    return new_image, new_bboxes, new_labels


def _get_flip_left_right_mats(
        convert_fn: Callable[..., T]
) -> Tuple[T, T]:
    image_dest_flip_lr_mat = convert_fn([
        [-1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ])
    bboxes_flip_lr_mat = convert_fn([
        [-1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]
    ])
    return image_dest_flip_lr_mat, bboxes_flip_lr_mat


def _flip_left_right(
        image: T,
        bboxes: T,
        labels: T,
        d: _AffineDependency
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image_dest_flip_lr_mat, bboxes_flip_lr_mat = _get_flip_left_right_mats(
        d.convert_fn
    )
    return _affine_transform(
        image, bboxes, labels, image_dest_flip_lr_mat, bboxes_flip_lr_mat,
        Interpolation.NEAREST, 0.0, d  # Dummy interpolation and fill_value.
    )


def _get_flip_up_down_mats(
        convert_fn: Callable[..., T]
) -> Tuple[T, T]:
    image_dest_flip_ud_mat = convert_fn([
        [1, 0, 0],
        [0, -1, 0],
        [0, 0, 1]
    ])
    bboxes_flip_ud_mat = convert_fn([
        [1, 0, 0],
        [0, -1, 0],
        [0, 0, 1]
    ])
    return image_dest_flip_ud_mat, bboxes_flip_ud_mat


def _flip_up_down(
        image: T,
        bboxes: T,
        labels: T,
        d: _AffineDependency
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    """
    image_dest_flip_ud_mat, bboxes_flip_ud_mat = _get_flip_up_down_mats(
        d.convert_fn
    )
    return _affine_transform(
        image, bboxes, labels, image_dest_flip_ud_mat, bboxes_flip_ud_mat,
        Interpolation.NEAREST, 0.0, d  # Dummy interpolation and fill_value.
    )


def _get_rotate_mats(
        angle_deg: T,
        convert_fn: Callable[..., T],
        cos_fn: Callable[[T], T],
        sin_fn: Callable[[T], T]
) -> Tuple[T, T]:
    ang_rad = convert_fn(np.pi * angle_deg / 180.0)
    # Image rotation matrix. Clockwise for the destination indices,
    # so the final image would appear to be rotated anti-clockwise.
    image_dest_rot_mat = convert_fn([
        [cos_fn(ang_rad), -sin_fn(ang_rad), 0],
        [sin_fn(ang_rad), cos_fn(ang_rad), 0],
        [0, 0, 1]
    ])
    bboxes_rot_mat = convert_fn([  # Anti-clockwise.
        [cos_fn(ang_rad), sin_fn(ang_rad), 0],
        [-sin_fn(ang_rad), cos_fn(ang_rad), 0],
        [0, 0, 1]
    ])
    return image_dest_rot_mat, bboxes_rot_mat


def _rotate(
        image: T,
        bboxes: T,
        labels: T,
        angle_deg: T,
        cos_fn: Callable[[T], T],
        sin_fn: Callable[[T], T],
        interpolation: Interpolation,
        fill_value: float,
        d: _AffineDependency
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    angle_deg: positive means anti-clockwise.
    """
    image_dest_rot_mat, bboxes_rot_mat = _get_rotate_mats(
        angle_deg, d.convert_fn, cos_fn, sin_fn
    )
    return _affine_transform(
        image, bboxes, labels, image_dest_rot_mat, bboxes_rot_mat,
        interpolation, fill_value, d
    )


def _get_shear_mats(
        angle_deg: T,
        convert_fn: Callable[..., T],
        tan_fn: Callable[[T], T]
) -> Tuple[T, T]:
    ang_rad = convert_fn(np.pi * angle_deg / 180.0)
    factor = tan_fn(ang_rad)
    # Image shear matrix. Clockwise for the destination indices,
    # so the final image would appear to be sheared anti-clockwise.
    image_dest_shear_mat = convert_fn([
        [1, -factor, 0],
        [0, 1, 0],
        [0, 0, 1]
    ])
    bboxes_shear_mat = convert_fn([  # Anti-clockwise.
        [1, factor, 0],
        [0, 1, 0],
        [0, 0, 1]
    ])
    return image_dest_shear_mat, bboxes_shear_mat


def _shear(
        image: T,
        bboxes: T,
        labels: T,
        angle_deg: T,
        tan_fn: Callable[[T], T],
        interpolation: Interpolation,
        fill_value: float,
        d: _AffineDependency
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    angle_deg: positive means anti-clockwise, where abs(angle_deg) must be < 90.
    """
    image_dest_shear_mat, bboxes_shear_mat = _get_shear_mats(
        angle_deg, d.convert_fn, tan_fn
    )
    return _affine_transform(
        image, bboxes, labels, image_dest_shear_mat, bboxes_shear_mat,
        interpolation, fill_value, d
    )


def _get_translate_mats(
        translate_height: T,
        translate_width: T,
        convert_fn: Callable[..., T]
) -> Tuple[T, T]:
    image_dest_translate_mat = convert_fn([
        [1, 0, -translate_width],
        [0, 1, -translate_height],
        [0, 0, 1]
    ])
    bboxes_translate_mat = convert_fn([
        [1, 0, translate_width],
        [0, 1, translate_height],
        [0, 0, 1]
    ])
    return image_dest_translate_mat, bboxes_translate_mat


def _translate(
        image: T,
        bboxes: T,
        labels: T,
        translate_height: T,
        translate_width: T,
        interpolation: Interpolation,
        fill_value: float,
        d: _AffineDependency
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    translate_height: in range (-image_height, image_height)
    translate_width: in range (-image_width, image_width)
    """
    image_dest_translate_mat, bboxes_translate_mat = _get_translate_mats(
        translate_height, translate_width, d.convert_fn
    )
    return _affine_transform(
        image, bboxes, labels, image_dest_translate_mat, bboxes_translate_mat,
        interpolation, fill_value, d
    )


def _get_random_size_fractions(
        height_fraction_range_: Tuple[float, float],
        width_fraction_range_: Tuple[float, float],
        rand_fn: Callable[..., T],
        convert_fn: Callable[..., T],
) -> Tuple[T, T]:
    """
    height_fraction_range, width_fraction_range: (-1.0, 1.0)
    rand_fn: generate random number in range [0.0, 1.0)
    """
    height_fraction_range = convert_fn(height_fraction_range_)
    width_fraction_range = convert_fn(width_fraction_range_)

    min_height_fraction = height_fraction_range[0]
    min_width_fraction = width_fraction_range[0]
    height_fraction_diff = height_fraction_range[1] - min_height_fraction
    width_fraction_diff = width_fraction_range[1] - min_width_fraction

    height_fraction = height_fraction_diff * rand_fn() + min_height_fraction
    width_fraction = width_fraction_diff * rand_fn() + min_width_fraction

    return height_fraction, width_fraction


def _get_crop_inputs(
        image_height_: int,
        image_width_: int,
        height_fraction_range: Tuple[float, float],
        width_fraction_range: Tuple[float, float],
        rand_fn: Callable[..., T],
        convert_fn: Callable[..., T],
        round_to_int_fn: Callable[[T], T]
) -> Tuple[T, T, T, T]:
    """
    height_fraction_range, width_fraction_range: in range [0.0, 1.0)
    rand_fn: generate random number in range [0.0, 1.0)
    """
    image_height = convert_fn(image_height_)
    image_width = convert_fn(image_width_)

    height_fraction, width_fraction = _get_random_size_fractions(
        height_fraction_range, width_fraction_range, rand_fn, convert_fn
    )

    crop_height: T = image_height * height_fraction
    crop_width: T = image_width * width_fraction

    offset_height = round_to_int_fn((image_height - crop_height) * rand_fn())
    offset_width = round_to_int_fn((image_width - crop_width) * rand_fn())

    return offset_height, offset_width, crop_height, crop_width


def _crop(
        image: T,
        bboxes: T,
        labels: T,
        offset_height: T,
        offset_width: T,
        crop_height: T,
        crop_width: T,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Sequence[int]],
        reshape_fn: Callable[[T, Sequence[int]], T],
        concat_fn: Callable[[List[T], int], T],
        logical_and_fn: Callable[[T, T], T],
        squeeze_fn: Callable[[T, int], T],
        clip_fn: Callable[[T, T, T], T],
        boolean_mask_fn: Callable[[T, T], T],
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    offset_height: in range [0, image_height)
    offset_width: in range [0, image_width)
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, shape_fn, reshape_fn
    )

    top = offset_height
    left = offset_width
    bottom: T = top + crop_height
    right: T = left + crop_width

    # Crop image.
    image = image[int(top):int(bottom), int(left):int(right), :]

    # Translate bboxes values.
    xs: T = bboxes[:, :1] - offset_width
    ys: T = bboxes[:, 1:2] - offset_height
    widths = bboxes[:, 2:3]
    heights = bboxes[:, 3:]

    # Filter bboxes values.
    xcens = xs + widths // 2
    ycens = ys + heights // 2
    included = squeeze_fn(logical_and_fn(
        logical_and_fn(xcens >= 0, xcens <= crop_width),
        logical_and_fn(ycens >= 0, ycens <= crop_height)
    ), -1)  # Squeeze along the last axis.

    # Clip bboxes values.
    xmaxs = clip_fn(xs + widths, convert_fn(0), crop_width)
    ymaxs = clip_fn(ys + heights, convert_fn(0), crop_height)
    xs = clip_fn(xs, convert_fn(0), crop_width)
    ys = clip_fn(ys, convert_fn(0), crop_height)
    widths = xmaxs - xs
    heights = ymaxs - ys

    bboxes = concat_fn([xs, ys, widths, heights], 1)
    bboxes = boolean_mask_fn(bboxes, included)

    # Filter labels.
    labels = boolean_mask_fn(labels, included)

    return image, bboxes, labels


def _resize(
        image: T,
        bboxes: T,
        labels: T,
        dest_size: Tuple[int, int],
        interpolation: Interpolation,
        convert_fn: Callable[..., T],
        shape_fn: Callable[[T], Sequence[int]],
        reshape_fn: Callable[[T, Sequence[int]], T],
        resize_image_fn: Callable[[T, Tuple[int, int], Interpolation], T],
        concat_fn: Callable[[List[T], int], T],
) -> Tuple[T, T, T]:
    """
    image: [h, w, c]
    bboxes: [[top_left_x, top_left_y, width, height], ...]
    labels: [0, 1, 0, ...]
    dest_size: (image_height, image_width)
    """
    image, bboxes, labels = _sanitise(
        image, bboxes, labels, convert_fn, shape_fn, reshape_fn
    )
    image_shape = shape_fn(image)

    image = resize_image_fn(image, dest_size, interpolation)

    w = convert_fn(dest_size[1] / image_shape[1])
    h = convert_fn(dest_size[0] / image_shape[0])

    xs = bboxes[:, :1] * w
    ys = bboxes[:, 1:2] * h
    widths = bboxes[:, 2:3] * w
    heights = bboxes[:, 3:] * h
    bboxes = concat_fn([xs, ys, widths, heights], 1)

    return image, bboxes, labels



#!/usr/bin/env python3
"""
TensorFlow Dataset test.
"""
from typing import Sequence, Tuple

import numpy as np
import numpy.typing

import targetran.tf

NDAnyArray = np.typing.NDArray[np.float_]


def make_np_data() -> Tuple[Sequence[NDAnyArray],
                            Sequence[NDAnyArray],
                            Sequence[NDAnyArray]]:
    image_seq = [np.random.rand(480, 512, 3) for _ in range(3)]

    bboxes_seq = [
        np.array([
            [214, 223, 10, 11],
            [345, 230, 21, 9],
        ]),
        np.array([]),
        np.array([
            [104, 151, 22, 10],
            [99, 132, 20, 15],
            [340, 220, 31, 12],
        ]),
    ]

    labels_seq = [
        np.array([0, 1]),
        np.array([]),
        np.array([2, 3, 0]),
    ]

    return image_seq, bboxes_seq, labels_seq


def main() -> None:
    image_seq, bboxes_seq, labels_seq = make_np_data()

    ds = targetran.tf.to_tf_dataset(image_seq, bboxes_seq, labels_seq)

    print("-------- Raw data --------")

    for example in ds:
        image, bboxes, labels = example
        print(f"image shape: {image.get_shape()}")
        print(f"bboxes shape: {bboxes.get_shape()}")
        print(f"labels shape: {labels.get_shape()}")
        print("=========")

    print("-------- Random transform --------")

    ds = ds \
        .map(targetran.tf.TFRandomRotate(probability=1.0)) \
        .map(targetran.tf.TFRandomShear(probability=1.0)) \
        .map(targetran.tf.TFRandomTranslate(probability=1.0)) \
        .map(targetran.tf.TFRandomFlipUpDown(probability=1.0)) \
        .map(targetran.tf.TFRandomFlipLeftRight(probability=1.0)) \
        .map(targetran.tf.TFRandomCrop(probability=1.0))

    for example in ds:
        image, bboxes, labels = example
        print(f"transformed image shape: {image.get_shape()}")
        print(f"transformed bboxes shape: {bboxes.get_shape()}")
        print(f"transformed bboxes: {bboxes.numpy().tolist()}")
        print(f"transformed labels shape: {labels.get_shape()}")
        print(f"transformed labels: {labels.numpy().tolist()}")
        print("=========")

    print("-------- Random transform with combine-affine --------")

    ds = targetran.tf.to_tf_dataset(image_seq, bboxes_seq, labels_seq)

    affine_transforms = targetran.tf.TFCombineAffine([
        targetran.tf.TFRandomRotate(probability=1.0),
        targetran.tf.TFRandomShear(probability=1.0),
        targetran.tf.TFRandomTranslate(probability=1.0),
        targetran.tf.TFRandomFlipUpDown(probability=1.0),
        targetran.tf.TFRandomFlipLeftRight(probability=1.0),
    ], probability=1.0)

    ds = ds \
        .map(targetran.tf.TFRandomCrop(probability=1.0)) \
        .map(affine_transforms) \
        .map(targetran.tf.TFResize((256, 256)))

    for example in ds:
        image, bboxes, labels = example
        print(f"transformed image shape: {image.get_shape()}")
        print(f"transformed bboxes shape: {bboxes.get_shape()}")
        print(f"transformed bboxes: {bboxes.numpy().tolist()}")
        print(f"transformed labels shape: {labels.get_shape()}")
        print(f"transformed labels: {labels.numpy().tolist()}")
        print("=========")

    print("-------- Batching --------")

    ds = ds.padded_batch(2, padding_values=np.nan)

    for batch in ds:
        image_batch, bboxes_batch, labels_batch = batch
        print(f"transformed image batch shape: {image_batch.get_shape()}")
        print(f"transformed bboxes batch shape: {bboxes_batch.get_shape()}")
        print(f"transformed bboxes batch: {bboxes_batch.numpy().tolist()}")
        print(f"transformed labels batch shape: {labels_batch.get_shape()}")
        print(f"transformed labels batch: {labels_batch.numpy().tolist()}")
        print("=========")


if __name__ == "__main__":
    main()



"""
TensorFlow functional helper utilities.
"""
from typing import Tuple

import tensorflow as tf  # type: ignore

from targetran._typing import ArrayLike
from targetran.utils import Interpolation

_TF_INTERPOLATION_DICT = {
    Interpolation.NEAREST: tf.image.ResizeMethod.NEAREST_NEIGHBOR,
    Interpolation.BILINEAR: tf.image.ResizeMethod.BILINEAR,
    Interpolation.BICUBIC: tf.image.ResizeMethod.BICUBIC,
    Interpolation.AREA: tf.image.ResizeMethod.AREA,
}


def _tf_convert(x: ArrayLike) -> tf.Tensor:
    if isinstance(x, tf.Tensor):
        return tf.cast(x, tf.float32)
    return tf.convert_to_tensor(x, tf.float32)


def _tf_cast_to_int(x: tf.Tensor) -> tf.Tensor:
    return tf.cast(x, tf.int32)


def _tf_round_to_int(x: tf.Tensor) -> tf.Tensor:
    return tf.cast(tf.math.rint(x), tf.int32)


def _tf_pad_image(
        image: tf.Tensor,
        pad_offsets: tf.Tensor,
) -> tf.Tensor:
    """
    pad_offsets: [top, bottom, left, right]
    """
    height = int(tf.shape(image)[0])
    width = int(tf.shape(image)[1])
    target_height = int(pad_offsets[0]) + height + int(pad_offsets[1])
    target_width = int(pad_offsets[2]) + width + int(pad_offsets[3])
    return tf.image.pad_to_bounding_box(
        image,
        int(pad_offsets[0]), int(pad_offsets[2]),
        target_height, target_width
    )


def _tf_resize_image(
        image: tf.Tensor,
        dest_size: Tuple[int, int],
        interpolation: Interpolation
) -> tf.Tensor:
    """
    dest_size: (image_height, image_width)
    """
    return tf.image.resize(
        image, size=dest_size, method=_TF_INTERPOLATION_DICT[interpolation]
    )


def _tf_gather_image(image: tf.Tensor, indices: tf.Tensor) -> tf.Tensor:
    """
    indices: [[row_0, row_1, ...], [col_0, col_1, ...]]
    """
    return tf.gather_nd(image, tf.transpose(indices))



"""
Sanity checks.
"""
from typing import Optional, Sequence, Tuple


def _check_shear_input(angle_deg: float) -> None:
    if not abs(angle_deg) < 90.0:
        raise ValueError(
            "The shear angle_deg has to be inside the open interval "
            "(-90.0, 90.0)."
        )


def _check_translate_input(
        image_shape: Sequence[int],
        translate_height: int,
        translate_width: int
) -> None:
    assert len(image_shape) == 3
    height_cond = -image_shape[0] < translate_height < image_shape[0]
    width_cond = -image_shape[1] < translate_width < image_shape[1]
    if not height_cond:
        raise ValueError(
            "The translate_height has to be inside the open interval "
            "(-image_height, image_height). In this case that means "
            f"{-image_shape[0]} < translate_height < {image_shape[0]}."
        )
    if not width_cond:
        raise ValueError(
            "The translate_width has to be inside the open interval "
            "(-image_width, image_width). In this case that means "
            f"{-image_shape[1]} < translate_width < {image_shape[1]}."
        )


def _check_crop_input(
        image_shape: Sequence[int],
        offset_height: int,
        offset_width: int
) -> None:
    assert len(image_shape) == 3
    height_cond = 0 <= offset_height < image_shape[0]
    width_cond = 0 <= offset_width < image_shape[1]
    if not height_cond:
        raise ValueError(
            "The offset_height has to be inside the half-open interval "
            "[0, image_height). In this case that means "
            f"0 <= offset_height < {image_shape[0]}."
        )
    if not width_cond:
        raise ValueError(
            "The translate_width has to be inside the half-open interval "
            "[0, image_width). In this case that means "
            f"0 <= offset_width < {image_shape[1]}."
        )


def _check_input_range(
        input_range: Tuple[float, float],
        limit_open_interval: Optional[Tuple[float, float]],
        input_name: str
) -> None:
    if limit_open_interval is None:
        if not input_range[0] <= input_range[1]:
            raise ValueError(
                f"The {input_name} should be provided as "
                f"(min_value, max_value), where min_value <= max_value."
            )
        return
    lower_limit, upper_limit = limit_open_interval
    if not lower_limit < input_range[0] <= input_range[1] < upper_limit:
        raise ValueError(
            f"The {input_name} should be provided as "
            f"(min_value, max_value), "
            f"where {lower_limit} < min_value <= max_value < {upper_limit}."
        )



"""
Utilities.
"""
from enum import auto, Enum
from typing import Any, Callable, Sequence, Tuple, Union

from targetran._typing import T


class Interpolation(Enum):
    NEAREST = auto()
    BILINEAR = auto()
    BICUBIC = auto()
    AREA = auto()


class Compose:

    def __init__(self, fns: Sequence[Callable[..., Any]]) -> None:
        self.fns = fns

    def __call__(self, *args: Any) -> Any:
        # This is needed, otherwise *args will "unpack" an image array.
        if len(args) == 1:
            x = args[0]
            for fn in self.fns:
                x = fn(x)
            return x

        for fn in self.fns:
            args = fn(*args)
        return args


def collate_fn(batch: Sequence[Tuple[Any, ...]]) -> Tuple[Sequence[Any], ...]:
    return tuple(zip(*batch))


def image_only(
        tran_fn: Callable[[T, Any, Any], Tuple[T, T, T]]
) -> Callable[..., Union[T, Tuple[T, ...]]]:
    # Only the image will be transformed.
    def fn(image: T, *args: Any) -> Any:
        transformed_image = tran_fn(image, [], [])[0]
        if len(args) == 0:
            return transformed_image
        return (transformed_image, *args)  # The parentheses are needed.

    return fn



#!/usr/bin/env python3
"""
TensorFlow Dataset timing.
"""
import os
from timeit import default_timer as timer
from typing import Iterator, Tuple

import tensorflow as tf  # type: ignore

from targetran.tf import (
    TFCombineAffine,
    TFRandomFlipLeftRight,
    TFRandomRotate,
    TFRandomShear,
    TFRandomCrop,
    TFRandomTranslate,
    TFResize,
)

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
AUTO = tf.data.AUTOTUNE
rng = tf.random.Generator.from_seed(42)


def generator() -> Iterator[Tuple[tf.Tensor, tf.Tensor, tf.Tensor]]:
    """
    Generate random data.
    """
    sample_size = 100000
    for _ in range(sample_size):
        height = rng.uniform(
            shape=(), minval=512, maxval=1024, dtype=tf.int32
        )
        width = rng.uniform(
            shape=(), minval=512, maxval=1024, dtype=tf.int32
        )
        image = rng.uniform(
            shape=(height, width, 3), minval=0, maxval=255, dtype=tf.int32
        )

        num_bboxes = rng.uniform(shape=(), minval=0, maxval=10, dtype=tf.int32)
        bboxes = rng.uniform(
            shape=(num_bboxes, 4), minval=16, maxval=256, dtype=tf.int32
        )
        labels = rng.uniform(
            shape=(num_bboxes,), minval=0, maxval=20, dtype=tf.int32
        )
        yield image, bboxes, labels


def main() -> None:
    ds = tf.data.Dataset.from_generator(
        generator,
        output_signature=(
            tf.TensorSpec((None, None, 3)),
            tf.TensorSpec((None, 4)),
            tf.TensorSpec((None,))
        )
    )

    affine_transform = TFCombineAffine([
        TFRandomFlipLeftRight(),
        TFRandomRotate(),
        TFRandomShear(),
        TFRandomTranslate()
    ])

    ds = ds \
        .map(TFRandomCrop(), num_parallel_calls=AUTO) \
        .map(affine_transform, num_parallel_calls=AUTO) \
        .map(TFResize(dest_size=(256, 256)), num_parallel_calls=AUTO)

    logging_batch_size = 100
    count = 0
    print("Start...")
    start = timer()
    total_start = start
    for _ in ds:
        count += 1
        if count % logging_batch_size == 0:
            print(f"- Runtime for recent {logging_batch_size} examples: "
                  f"{timer() - start} s; "
                  f"total number of examples so far: {count}")
            start = timer()
    print("--------------")
    print(f"Total runtime for {count} examples: {timer() - total_start}")


if __name__ == "__main__":
    main()



"""
Numpy functional helper utilities.
"""
from typing import Tuple

import cv2
import numpy as np

from targetran._typing import (
    ArrayLike,
    NDAnyArray,
    NDFloatArray,
    NDIntArray,
    NDBoolArray,
)
from targetran.utils import Interpolation

_INTERPOLATION_DICT = {
    Interpolation.NEAREST: cv2.INTER_NEAREST,  # pylint: disable=no-member
    Interpolation.BILINEAR: cv2.INTER_LINEAR,  # pylint: disable=no-member
    Interpolation.BICUBIC: cv2.INTER_CUBIC,  # pylint: disable=no-member
    Interpolation.AREA: cv2.INTER_AREA,  # pylint: disable=no-member
}


def _np_convert(x: ArrayLike) -> NDFloatArray:
    if isinstance(x, np.ndarray):
        return x.astype(np.float32)
    return np.array(x, dtype=np.float32)


def _np_range(start: int, end: int, step: int) -> NDIntArray:
    return np.arange(start, end, step, dtype=np.int32)


def _np_cast_to_int(x: NDAnyArray) -> NDIntArray:
    return x.astype(dtype=np.int32)


def _np_round_to_int(x: NDAnyArray) -> NDIntArray:
    return np.rint(x.astype(dtype=np.float32)).astype(dtype=np.int32)


def _np_logical_and(x: NDBoolArray, y: NDBoolArray) -> NDBoolArray:
    return np.logical_and(x, y)


def _np_pad_image(
        image: NDFloatArray,
        pad_offsets: NDIntArray
) -> NDFloatArray:
    """
    pad_offsets: [top, bottom, left, right]
    """
    pad_width = (  # From axis 0 to 2.
        (int(pad_offsets[0]), int(pad_offsets[1])),
        (int(pad_offsets[2]), int(pad_offsets[3])),
        (0, 0)
    )
    return np.pad(image, pad_width=pad_width, constant_values=0)


def _np_resize_image(
        image: NDFloatArray,
        dest_size: Tuple[int, int],
        interpolation: Interpolation
) -> NDFloatArray:
    """
    dest_size: (image_height, image_width)
    """
    resized_image: NDAnyArray = cv2.resize(  # pylint: disable=no-member
        image,
        dsize=(dest_size[1], dest_size[0]),
        interpolation=_INTERPOLATION_DICT[interpolation]
    )
    return resized_image


def _np_boolean_mask(x: NDAnyArray, mask: NDBoolArray) -> NDAnyArray:
    """
    mask: boolean array
    """
    return x[mask]


def _np_gather_image(image: NDFloatArray, indices: NDIntArray) -> NDFloatArray:
    """
    indices: [[row_0, row_1, ...], [col_0, col_1, ...]]
    """
    return image[tuple(indices)]  # type: ignore



from ._np import (
    flip_left_right as flip_left_right,
    flip_up_down as flip_up_down,
    rotate as rotate,
    shear as shear,
    translate as translate,
    crop as crop,
    resize as resize,
    CombineAffine as CombineAffine,
    RandomTransform as RandomTransform,
    RandomFlipLeftRight as RandomFlipLeftRight,
    RandomFlipUpDown as RandomFlipUpDown,
    RandomRotate as RandomRotate,
    RandomShear as RandomShear,
    RandomTranslate as RandomTranslate,
    RandomCrop as RandomCrop,
    Resize as Resize,
)
